import unittest
from enum import Enum
import math
import cbrrr

class MajorType(Enum):
	UNSIGNED_INT = 0
	NEGATIVE_INT = 1
	BYTE_STRING = 2
	TEXT_STRING = 3
	ARRAY = 4
	MAP = 5
	TAG = 6
	FLOAT = 7


def cbor_head(mtype, info):
	if type(mtype) is MajorType:
		mtype = mtype.value
	assert(mtype < 8)
	assert(info < 32)
	return bytes([mtype << 5 | info])


def roundrip(obj, atjson_mode=False):
	return cbrrr.decode_dag_cbor( cbrrr.encode_dag_cbor(obj, atjson_mode) )

class TestCbrrr(unittest.TestCase):
	def test_simple_roundtrips(self):
		self.assertEqual(roundrip(123), 123)
		self.assertEqual(roundrip(-123), -123)
		self.assertEqual(roundrip(3.1415), 3.1415)
		self.assertEqual(roundrip("hello"), "hello")
		self.assertEqual(roundrip(b"world"), b"world")
		self.assertEqual(roundrip(True), True)
		self.assertEqual(roundrip(False), False)
		self.assertEqual(roundrip(None), None)
		self.assertEqual(roundrip(cbrrr.CID(b"\x01q\x12 "+b"A"*32)), cbrrr.CID(b"\x01q\x12 "+b"A"*32))
		self.assertEqual(roundrip([1, 2, 3]), [1, 2, 3])
		self.assertEqual(roundrip({"a": 1, "b": 2}), {"a": 1, "b": 2})
		self.assertEqual(roundrip(0xffffffffffffffff), 0xffffffffffffffff)
		self.assertEqual(roundrip(~0xffffffffffffffff), ~0xffffffffffffffff)
	
	def test_decode_small_integers(self):
		self.assertEqual(cbrrr.decode_dag_cbor(cbor_head(MajorType.UNSIGNED_INT, 0)), 0)
		self.assertEqual(cbrrr.decode_dag_cbor(cbor_head(MajorType.UNSIGNED_INT, 1)), 1)
		self.assertEqual(cbrrr.decode_dag_cbor(cbor_head(MajorType.UNSIGNED_INT, 22)), 22)
		self.assertEqual(cbrrr.decode_dag_cbor(cbor_head(MajorType.UNSIGNED_INT, 23)), 23)
	
	def test_encode_toobig_integers(self):
		self.assertRaises(ValueError, cbrrr.encode_dag_cbor, 0xffffffffffffffff + 1)
		self.assertRaises(ValueError, cbrrr.encode_dag_cbor, ~(0xffffffffffffffff + 1))
	
	def test_encode_illegal_floats(self):
		self.assertRaises(ValueError, cbrrr.encode_dag_cbor, math.nan)
		self.assertRaises(ValueError, cbrrr.encode_dag_cbor, math.inf)
		self.assertRaises(ValueError, cbrrr.encode_dag_cbor, -math.inf)

	def test_decode_invalid_small_integers(self):
		for i in range(24, 28):
			self.assertRaises(cbrrr.CbrrrDecodeError, cbrrr.decode_dag_cbor, cbor_head(MajorType.UNSIGNED_INT, i))
		for i in range(28, 32):
			self.assertRaises(cbrrr.CbrrrDecodeError, cbrrr.decode_dag_cbor, cbor_head(MajorType.UNSIGNED_INT, i))
	
	def test_atjson_decode(self):
		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(cbrrr.CID(b"blah")), atjson_mode=True), {'$link': 'bmjwgc2a'})

		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(b"hello"), atjson_mode=True), {'$bytes': 'aGVsbG8'})
		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(b""),      atjson_mode=True), {'$bytes': ''})
		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(b"A"),     atjson_mode=True), {'$bytes': 'QQ'})
		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(b"AB"),    atjson_mode=True), {'$bytes': 'QUI'})
		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(b"ABC"),   atjson_mode=True), {'$bytes': 'QUJD'})
		self.assertEqual(cbrrr.decode_dag_cbor(cbrrr.encode_dag_cbor(b"ABCD"),  atjson_mode=True), {'$bytes': 'QUJDRA'})

	def test_atjson_encode_rejects_bytes_and_cid(self):
		self.assertRaises(TypeError, cbrrr.encode_dag_cbor, b"hello", True)
		self.assertRaises(TypeError, cbrrr.encode_dag_cbor, cbrrr.CID(b"blah"), True)

	def test_multi_decode(self):
		self.assertEqual(
			list(cbrrr.decode_multi_dag_cbor_in_violation_of_the_spec(
				cbrrr.encode_dag_cbor(b"hello") +
				cbrrr.encode_dag_cbor({"world": 0}) +
				cbrrr.encode_dag_cbor([1, 2, 3])
			)),
			[b'hello', {"world": 0}, [1, 2, 3]]
		)
	
	def test_duplicate_map_keys(self):
		# {"abc": 1, "abc": 2}
		dup = cbor_head(MajorType.MAP, 2)
		dup += cbor_head(MajorType.TEXT_STRING, 3) + b"abc"
		dup += cbor_head(MajorType.UNSIGNED_INT, 1)
		dup += cbor_head(MajorType.TEXT_STRING, 3) + b"abc"
		dup += cbor_head(MajorType.UNSIGNED_INT, 2)

		with self.assertRaises(ValueError) as cm:
			cbrrr.decode_dag_cbor(dup)
	
		self.assertEqual(
			str(cm.exception),
			"non-canonical map key ordering ('abc' <= 'abc')"
		)
	
	def test_unsorted_map_keys(self):
		# {"def": 1, "abc": 2}
		obj = cbor_head(MajorType.MAP, 2)
		obj += cbor_head(MajorType.TEXT_STRING, 3) + b"def"
		obj += cbor_head(MajorType.UNSIGNED_INT, 1)
		obj += cbor_head(MajorType.TEXT_STRING, 3) + b"abc"
		obj += cbor_head(MajorType.UNSIGNED_INT, 2)

		with self.assertRaises(ValueError) as cm:
			cbrrr.decode_dag_cbor(obj)
	
		self.assertEqual(
			str(cm.exception),
			"non-canonical map key ordering ('abc' <= 'def')"
		)

		# {"aaa": 1, "x": 2}  (shorter string should sort first)
		obj = cbor_head(MajorType.MAP, 2)
		obj += cbor_head(MajorType.TEXT_STRING, 3) + b"aaa"
		obj += cbor_head(MajorType.UNSIGNED_INT, 1)
		obj += cbor_head(MajorType.TEXT_STRING, 1) + b"x"
		obj += cbor_head(MajorType.UNSIGNED_INT, 2)

		with self.assertRaises(ValueError) as cm:
			cbrrr.decode_dag_cbor(obj)
	
		self.assertEqual(
			str(cm.exception),
			"non-canonical map key ordering (len('x') < len('aaa'))"
		)

if __name__ == '__main__':
	unittest.main(module="tests.test_cbrrr")



from typing import Type, Iterator, Union, Callable, Any
import base64
import hashlib
from . import _cbrrr  # type: ignore

CbrrrDecodeError = _cbrrr.CbrrrDecodeError

class CID:
	"""
	This class is very minimal, intended to support atproto use cases and not
	much else.
	"""

	CIDV1_DAG_CBOR_SHA256_32_PFX = b"\x01\x71\x12\x20"
	CIDV1_RAW_SHA256_32_PFX      = b"\x01\x55\x12\x20"

	__slots__ = ("cid_bytes",)

	def __init__(self, cid_bytes: bytes) -> None:
		"""
		Expects raw byes, without a multibase prefix.

		If you don't have raw bytes, you probably want CID.decode()

		NOTE: No validation is performed here! You're responsible for ensuring
		the CID has a format you recognise. the is_cidv1_dag_cbor_sha256_32()
		and is_cidv1_raw_sha256_32() methods may be useful for this.
		"""
		self.cid_bytes = cid_bytes
	
	@classmethod
	def cidv1_dag_cbor_sha256_32_from(cls, data: bytes) -> "CID":
		return cls(
			cls.CIDV1_DAG_CBOR_SHA256_32_PFX
			+ hashlib.sha256(data).digest()
		)

	@classmethod
	def cidv1_raw_sha256_32_from(cls, data: bytes) -> "CID":
		return cls(
			cls.CIDV1_RAW_SHA256_32_PFX
			+ hashlib.sha256(data).digest()
		)
	
	@classmethod
	def decode(cls, data: Union[bytes, str]) -> "CID":
		"""
		Currently supported codecs: identity/raw, base32
		"""

		if type(data) is str:
			data = data.encode()

		if data.startswith(b"\x00"): # identity multibase codec
			return cls(data[1:])

		if data.startswith(b"b"): # base32 multibase codec
			data = data[1:] # strip prefix
			if data.endswith(b"="):
				raise ValueError("unexpected base32 padding")
			# add back correct amount of padding (python is fussy)
			data += b"=" * ((-len(data)) % 8)
			decoded = base64.b32decode(data, casefold=True)
			return cls(decoded)

		raise ValueError("I don't know how to decode this CID")

	def encode(self, base="base32") -> str:
		if base == "base32":
			return "b" + base64.b32encode(self.cid_bytes) \
				.decode().lower().rstrip("=")
		# this function might support other encodings in the future
		raise ValueError("unsupported base encoding")

	def is_cidv1_dag_cbor_sha256_32(self) -> bool:
		return (
			self.cid_bytes.startswith(self.CIDV1_DAG_CBOR_SHA256_32_PFX)
			and len(self.cid_bytes) == 36
		)

	def is_cidv1_raw_sha256_32(self) -> bool:
		return (
			self.cid_bytes.startswith(self.CIDV1_RAW_SHA256_32_PFX)
			and len(self.cid_bytes) == 36
		)

	def __bytes__(self):
		return self.cid_bytes

	def __repr__(self):
		return f"CID({self.encode()})"
	
	def __hash__(self) -> int:
		return self.cid_bytes.__hash__()
	
	def __eq__(self, __value: object) -> bool:
		if not isinstance(__value, CID):
			return False
		return self.cid_bytes == __value.cid_bytes

# nb: | syntax not supported in <=py3.9
DagCborTypes = Union[str, bytes, int, bool, float, CID, list, dict, None]

def decode_dag_cbor(
	data: bytes,
	atjson_mode: bool=False,
	cid_ctor: Callable[[bytes], Any]=CID
) -> DagCborTypes:
	"""
	Decode DAG-CBOR bytes into python objects.

	If atjson_mode is True, bytes will be represented as {"$bytes": "b64..."},
	and CIDs will be represented as {"$link": "b32..."}. Otherwise they'll
	be represented as bytes objects, or CID classes, respectively.
	"""

	parsed, length = _cbrrr.decode_dag_cbor(data, cid_ctor, atjson_mode)
	if length != len(data):
		raise ValueError("did not parse to end of buffer")
	return parsed

def decode_multi_dag_cbor_in_violation_of_the_spec(
	data: bytes,
	atjson_mode: bool=False,
	cid_ctor: Callable[[bytes], Any]=CID
) -> Iterator[DagCborTypes]:
	"""
	https://ipld.io/specs/codecs/dag-cbor/spec/#strictness

	"Encode and decode must operate on a single top-level CBOR object.
	Back-to-back concatenated objects are not allowed or supported, as suggested
	by section 5.1 of RFC 8949 for streaming applications."
	"""
	view = memoryview(data)
	offset = 0
	while offset < len(data):
		parsed, length = _cbrrr.decode_dag_cbor(view[offset:], cid_ctor, atjson_mode)
		yield parsed
		offset += length
	assert(offset == len(data)) # should never fail!

def encode_dag_cbor(
	obj: DagCborTypes,
	atjson_mode: bool=False,
	cid_type: Type=CID
) -> bytes:
	"""
	Encode python objects to DAG-CBOR bytes.

	If atjson_mode is True, dicts in the format {"$bytes": "b64..."} will be
	encoded as CBOR bytes, and dicts in the format {"$link": "b32..."} will be
	encoded as CIDs (CBOR tag value 42)
	"""
	return _cbrrr.encode_dag_cbor(obj, cid_type, atjson_mode)

__all__ = [
	"CbrrrDecodeError",
	"CID",
	"DagCborTypes",
	"decode_dag_cbor",
	"decode_multi_dag_cbor_in_violation_of_the_spec",
	"encode_dag_cbor",
]



from typing import Type, TypeVar, Tuple, Callable, Any

CbrrrDecodeErrorType = TypeVar("CbrrrDecodeErrorType", bound=ValueError)
CbrrrDecodeError: CbrrrDecodeErrorType

def decode_dag_cbor(buf: bytes, cid_ctor: Callable[[bytes], Any], atjson_mode: bool) -> Tuple[Any, int]:
	...

def encode_dag_cbor(obj: Any, cid_type: Type, atjson_mode: bool) -> bytes:
	...



# python3 setup.py build
# python3 setup.py develop --user

from setuptools import setup, Extension

setup(
	name="cbrrr",
	packages=["cbrrr"],
	package_dir={"cbrrr": "src/cbrrr"},
	ext_modules=[
		Extension(
			"cbrrr._cbrrr",
			sources=["src/cbrrr/_cbrrr.c"],
			extra_compile_args=["-O3", "-Wall", "-Wextra", "-Wpedantic", "-std=c99"]
		),
	]
)



import os
import cbrrr
import json
import base64
import base58
import unittest

class CID:
	def __init__(self, cid_bytes):
		self.cid_bytes = cid_bytes
	
	def __repr__(self) -> str:
		return f"CID({self.cid_bytes.hex()})"

	def __str__(self) -> str:
		#print(self.cid_bytes[:5])
		# XXX: this is a hack! do proper multiformat logic
		if self.cid_bytes.startswith(b'\x01'):
			return "b" + base64.b32encode(self.cid_bytes).decode().lower().rstrip("=")
		else:
			return base58.b58encode(self.cid_bytes).decode()

# XXX: this is not a good DAG-JSON encoder, it's just here because I need it
# to verify the test results
class DagJsonEncoder(json.JSONEncoder):
	def default(self, o):
		if isinstance(o, bytes):
			return {"/": {"bytes": base64.b64encode(o).decode().rstrip("=")}}
		if isinstance(o, CID):
			return {"/": str(o)}
		return json.JSONEncoder.default(self, o)


class TestFixtures(unittest.TestCase):
	FIXTURE_PATH = "codec-fixtures/fixtures/"

	def setUp(self):
		tests = []
		for subdir in os.listdir(self.FIXTURE_PATH):
			if subdir.startswith("."):
				continue

			dirpath = self.FIXTURE_PATH + subdir + "/"
			paths = os.listdir(dirpath)
			with open(next(dirpath + p for p in paths if p.endswith(".dag-cbor")), "rb") as infile:
				dag_cbor = infile.read()
			with open(next(dirpath + p for p in paths if p.endswith(".dag-json")), "rb") as outfile:
				dag_json = outfile.read()
			py_normalised_json = json.dumps(json.loads(dag_json)) # roundtrip thru python, this makes sure floats etc are in python-flavoured encodings
			tests.append((subdir, dag_cbor, py_normalised_json))
		self.tests = tests

	def test_all_fixtures(self):
		for name, cbor_in, expected in self.tests:
			result, parsed_len = cbrrr._cbrrr.decode_dag_cbor(cbor_in, CID, False)
			self.assertEqual(parsed_len, len(cbor_in))
			reserialised = json.dumps(result, cls=DagJsonEncoder, sort_keys=True)
			self.assertEqual(reserialised, expected, name)


if __name__ == '__main__':
	unittest.main(module="tests.test_fixtures")
